from django.db.models.signals import post_delete
from django.dispatch import receiver


@receiver(post_delete, weak=False, dispatch_uid="edc_reference_post_delete")
def reference_post_delete(instance, using, **kwargs):
    try:
        instance.reference_deleter_cls(model_obj=instance)
    except AttributeError:
        pass



from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, tag

from edc_base.utils import get_utcnow

from ..models import Reference
from ..populater import Populater
from ..reference_model_config import ReferenceModelConfig
from ..site import site_reference_configs
from .models import SubjectVisit, CrfOne


class TestPopulater(TestCase):

    def setUp(self):
        self.subject_identifier = '12345'
        site_reference_configs.registry = {}
        site_reference_configs.loaded = False
        reference = ReferenceModelConfig(
            name='edc_reference.subjectvisit',
            fields=['report_datetime', 'visit_code'])
        site_reference_configs.register(reference=reference)
        reference = ReferenceModelConfig(
            name='edc_reference.crfone',
            fields=['field_date', 'field_datetime', 'field_int', 'field_str'])
        site_reference_configs.register(reference=reference)
        self.report_datetime = get_utcnow()
        self.report_date = get_utcnow().date()
        self.subject_visit1 = SubjectVisit.objects.create(
            subject_identifier=self.subject_identifier,
            report_datetime=self.report_datetime)
        CrfOne.objects.create(
            subject_visit=self.subject_visit1,
            field_date=self.report_date,
            field_datetime=self.report_datetime,
            field_int=1,
            field_str='erik')
        self.subject_visit2 = SubjectVisit.objects.create(
            subject_identifier=self.subject_identifier,
            report_datetime=self.report_datetime - relativedelta(years=1))
        CrfOne.objects.create(
            subject_visit=self.subject_visit2,
            field_date=self.report_date,
            field_datetime=self.report_datetime,
            field_int=1,
            field_str='erik')

    def test_populates_for_visit(self):
        Reference.objects.all().delete()
        populater = Populater()
        populater.populate()
        for visit in SubjectVisit.objects.all():
            with self.subTest(report_datetime=visit.report_datetime):
                try:
                    Reference.objects.get(
                        identifier=self.subject_identifier,
                        model='edc_reference.subjectvisit',
                        report_datetime=visit.report_datetime,
                        field_name='report_datetime',
                        value_datetime=visit.report_datetime)
                except ObjectDoesNotExist as e:
                    self.fail(f'Object unexpectedly DoesNotExist. Got {e}')

    def test_populates_for_crfone_field_date(self):
        Reference.objects.all().delete()
        populater = Populater()
        populater.populate()
        for visit in SubjectVisit.objects.all():
            with self.subTest(report_datetime=visit.report_datetime):
                try:
                    Reference.objects.get(
                        identifier=self.subject_identifier,
                        model='edc_reference.crfone',
                        report_datetime=visit.report_datetime,
                        field_name='field_date',
                        value_date=self.report_date)
                except ObjectDoesNotExist as e:
                    self.fail(f'Object unexpectedly DoesNotExist. Got {e}')

    def test_populates_for_crfone_field_datetime(self):
        Reference.objects.all().delete()
        populater = Populater()
        populater.populate()
        for visit in SubjectVisit.objects.all():
            with self.subTest(report_datetime=visit.report_datetime):
                try:
                    Reference.objects.get(
                        identifier=self.subject_identifier,
                        model='edc_reference.crfone',
                        report_datetime=visit.report_datetime,
                        field_name='field_datetime',
                        value_datetime=self.report_datetime)
                except ObjectDoesNotExist as e:
                    self.fail(f'Object unexpectedly DoesNotExist. Got {e}')

    def test_populates_for_crfone_field_str(self):
        Reference.objects.all().delete()
        populater = Populater()
        populater.populate()
        for visit in SubjectVisit.objects.all():
            with self.subTest(report_datetime=visit.report_datetime):
                try:
                    Reference.objects.get(
                        identifier=self.subject_identifier,
                        report_datetime=visit.report_datetime,
                        model='edc_reference.crfone',
                        field_name='field_str',
                        value_str='erik')
                except ObjectDoesNotExist as e:
                    self.fail(f'Object unexpectedly DoesNotExist. Got {e}')

    def test_populates_for_crfone_field_int(self):
        Reference.objects.all().delete()
        populater = Populater()
        populater.populate()
        for visit in SubjectVisit.objects.all():
            with self.subTest(report_datetime=visit.report_datetime):
                try:
                    Reference.objects.get(
                        identifier=self.subject_identifier,
                        model='edc_reference.crfone',
                        report_datetime=visit.report_datetime,
                        field_name='field_int',
                        value_int=1)
                except ObjectDoesNotExist as e:
                    self.fail(f'Object unexpectedly DoesNotExist. Got {e}')

    def test_populater_updates(self):
        populater = Populater()
        populater.populate()
        for visit in SubjectVisit.objects.all():
            with self.subTest(report_datetime=visit.report_datetime):
                try:
                    Reference.objects.get(
                        identifier=self.subject_identifier,
                        model='edc_reference.crfone',
                        report_datetime=visit.report_datetime,
                        field_name='field_int',
                        value_int=1)
                except ObjectDoesNotExist as e:
                    self.fail(f'Object unexpectedly DoesNotExist. Got {e}')



# Generated by Django 2.0.1 on 2018-01-16 13:28

from django.db import migrations, models
import django.db.models.deletion
import edc_base.sites.managers


class Migration(migrations.Migration):

    dependencies = [
        ('sites', '0002_alter_domain_unique'),
        ('edc_reference', '0003_reference_related_name'),
    ]

    operations = [
        migrations.AlterModelManagers(
            name='reference',
            managers=[
                ('on_site', edc_base.sites.managers.CurrentSiteManager()),
            ],
        ),
        migrations.AddField(
            model_name='reference',
            name='site',
            field=models.ForeignKey(editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, to='sites.Site'),
        ),
    ]



from django.db import models
from edc_base.model_mixins import BaseUuidModel
from edc_base.sites import CurrentSiteManager, SiteModelMixin

from .managers import ReferenceManager
from django.conf import settings


class ReferenceFieldDatatypeNotFound(Exception):
    pass


class Reference(SiteModelMixin, BaseUuidModel):

    identifier = models.CharField(max_length=50)

    timepoint = models.CharField(max_length=50)

    report_datetime = models.DateTimeField()

    model = models.CharField(max_length=250)

    field_name = models.CharField(max_length=50)

    datatype = models.CharField(max_length=50)

    value_str = models.CharField(max_length=50, null=True)

    value_int = models.IntegerField(null=True)

    value_date = models.DateField(null=True)

    value_datetime = models.DateTimeField(null=True)

    value_uuid = models.UUIDField(null=True)

    related_name = models.CharField(max_length=100, null=True)

    on_site = CurrentSiteManager()

    objects = ReferenceManager()

    def __str__(self):
        return (f'{self.identifier}@{self.timepoint} {self.model}.'
                f'{self.field_name}={self.value}')

    def natural_key(self):
        return (self.identifier, self.timepoint, self.report_datetime,
                self.model, self.field_name)
    natural_key.dependencies = ['sites.Site']

    def update_value(self, value=None, internal_type=None, field=None, related_name=None):
        """Updates the correct `value` field based on the
        field class datatype.
        """
        internal_type = internal_type or field.get_internal_type()
        if internal_type in ['ForeignKey', 'OneToOneField']:
            self.datatype = 'UUIDField'
        else:
            self.datatype = internal_type
        update = None
        value_fields = [fld for fld in self._meta.get_fields()
                        if fld.name.startswith('value')]
        for fld in value_fields:
            if fld.name.startswith('value'):  # e.g. value_str, value_int, etc
                if fld.get_internal_type() == self.datatype:
                    update = (fld.name, value)
                    break
        if update:
            self.related_name = related_name
            setattr(self, *update)
            for fld in value_fields:
                if fld.name != update[0]:
                    setattr(self, fld.name, None)
        else:
            raise ReferenceFieldDatatypeNotFound(
                f'Reference field internal_type not found. Got \'{self.datatype}\'. '
                f'model={self.model}.{self.field_name} '
                f'Expected a django.models.field internal type like \'CharField\', '
                '\'DateTimeField\', etc.')
        self.save()

    @property
    def value(self):
        for field_name in ['value_str', 'value_int', 'value_date',
                           'value_datetime', 'value_uuid']:
            value = getattr(self, field_name)
            if value is not None:
                break
        return value

    class Meta:
        unique_together = ['identifier', 'timepoint',
                           'report_datetime', 'model', 'field_name']
        index_together = ['identifier', 'timepoint',
                          'report_datetime', 'model', 'field_name']
        ordering = ('identifier', 'report_datetime')
        indexes = [
            models.Index(fields=['identifier', 'timepoint', 'model']),
            models.Index(
                fields=['identifier', 'report_datetime', 'timepoint', 'model']),
            models.Index(fields=['report_datetime', 'timepoint']),
        ]


if settings.APP_NAME == 'edc_reference':
    from .tests.models import *



from dateutil.relativedelta import relativedelta

from django.test import TestCase, tag

from edc_base.utils import get_utcnow

from ..models import Reference
from ..reference_model_config import ReferenceModelConfig
from ..refsets import Refset, RefsetError, RefsetOverlappingField
from ..site import site_reference_configs
from .models import SubjectVisit, CrfOne


class TestRefset(TestCase):

    def setUp(self):
        self.subject_identifier = '12345'
        site_reference_configs.registry = {}
        site_reference_configs.loaded = False
        reference = ReferenceModelConfig(
            name='edc_reference.subjectvisit',
            fields=['report_datetime', 'visit_code'])
        site_reference_configs.register(reference=reference)
        reference = ReferenceModelConfig(
            name='edc_reference.crfone',
            fields=['field_date', 'field_datetime', 'field_int', 'field_str'])
        site_reference_configs.register(reference=reference)

        values = [
            ('NEG', get_utcnow() - relativedelta(years=3)),
            ('POS', get_utcnow() - relativedelta(years=2)),
            ('POS', get_utcnow() - relativedelta(years=1)),
        ]
        self.subject_visits = []
        for index in [0, 1, 2]:
            self.subject_visits.append(
                SubjectVisit.objects.create(
                    subject_identifier=self.subject_identifier,
                    report_datetime=get_utcnow() - relativedelta(months=3 - index),
                    visit_code=str(index)))
        for index, subject_visit in enumerate(SubjectVisit.objects.filter(
                subject_identifier=self.subject_identifier).order_by('report_datetime')):
            CrfOne.objects.create(
                subject_visit=subject_visit,
                field_int=index,
                field_str=values[index][0],
                field_datetime=values[index][1],
                field_date=values[index][1].date())

    def test_raises_model_not_in_site(self):
        site_reference_configs.registry = {}
        self.assertRaises(
            RefsetError,
            Refset,
            name='edc_reference.blah',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)

    def test_raises_unknown_reference_model(self):
        self.assertRaises(
            RefsetError,
            Refset,
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=None)

    def test_raises_missing_report_datetime(self):
        self.assertRaises(
            RefsetError,
            Refset,
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=None,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)

    def test_raises_missing_subject_identifier(self):
        self.assertRaises(
            RefsetError,
            Refset,
            name='edc_reference.crfone',
            subject_identifier=None,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)

    def test_raises_missing_timepoint(self):
        self.assertRaises(
            RefsetError,
            Refset,
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=None,
            reference_model_cls=Reference)

    def test_no_reference_instance(self):
        Reference.objects.all().delete()
        refset = Refset(
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)
        for value in refset._fields.values():
            self.assertIsNone(value)

    def test_missing_reference_instance_for_one_field(self):
        Reference.objects.filter(field_name='field_str').delete()
        refset = Refset(
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)
        self.assertIsNone(refset._fields.get('field_str'))

    def test_if_reference_exists_updates_report_datetime_in_fields(self):
        refset = Refset(
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)
        for field, value in refset._fields.items():
            if field == 'report_datetime':
                self.assertEqual(value, self.subject_visits[0].report_datetime)
            elif field == 'visit_code':
                self.assertEqual(
                    value, self.subject_visits[0].visit_code, msg=field)

    def test_if_reference_updates_fields(self):
        for index, subject_visit in enumerate(self.subject_visits):
            with self.subTest(index=index, subject_visit=subject_visit):
                refset = Refset(
                    name='edc_reference.crfone',
                    subject_identifier=subject_visit.subject_identifier,
                    report_datetime=subject_visit.report_datetime,
                    timepoint=subject_visit.visit_code,
                    reference_model_cls=Reference)
                crf_one = CrfOne.objects.get(subject_visit=subject_visit)
                for field, value in refset._fields.items():
                    if field == 'report_datetime':
                        self.assertEqual(
                            value, subject_visit.report_datetime, msg=field)
                    elif field == 'visit_code':
                        self.assertEqual(
                            value, subject_visit.visit_code, msg=field)
                    else:
                        self.assertEqual(value, getattr(
                            crf_one, field), msg=field)

    def test_raises_on_overlapping_field(self):
        site_reference_configs.add_fields_to_config(
            'edc_reference.crfone', ['subject_identifier'])
        Reference.objects.create(
            model='edc_reference.crfone',
            identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            field_name='subject_identifier')
        self.assertRaises(
            RefsetOverlappingField,
            Refset,
            name='edc_reference.crfone',
            subject_identifier=self.subject_identifier,
            report_datetime=self.subject_visits[0].report_datetime,
            timepoint=self.subject_visits[0].visit_code,
            reference_model_cls=Reference)



import sys

from django.apps import AppConfig as DjangoAppConfig
from django.core.checks.registry import register

from .site import site_reference_configs
from .system_checks import check_site_reference_configs


class AppConfig(DjangoAppConfig):
    name = 'edc_reference'
    verbose_name = 'Edc Reference'

    def ready(self):
        from .signals import reference_post_delete
        sys.stdout.write(f'Loading {self.verbose_name} ...\n')

        site_reference_configs.autodiscover()

        sys.stdout.write(f' Done loading {self.verbose_name}.\n')
        register(check_site_reference_configs)



# Generated by Django 2.0.1 on 2018-01-13 09:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_reference', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='reference',
            name='value_uuid',
            field=models.UUIDField(null=True),
        ),
    ]



from django.views.generic.base import RedirectView
from django.urls.conf import path

from .admin_site import edc_reference_admin

app_name = 'edc_reference'

urlpatterns = [
    path('admin/', edc_reference_admin.urls),
    path('', RedirectView.as_view(url='admin/edc_reference/'), name='home_url'),
]



from django.test import TestCase, tag
from edc_visit_schedule.site_visit_schedules import site_visit_schedules

from ..reference_model_config import ReferenceModelConfig
from ..reference_model_config import ReferenceFieldValidationError, ReferenceDuplicateField
from ..reference_model_config import ReferenceModelValidationError
from ..site import site_reference_configs
from ..site import AlreadyRegistered
from ..site import SiteReferenceConfigError, ReferenceConfigNotRegistered
from .visit_schedule import visit_schedule


class TestSite(TestCase):

    def setUp(self):
        site_visit_schedules._registry = {}
        site_visit_schedules.register(visit_schedule)

    def test_site_register(self):
        name = 'edc_reference.crfone'
        fields = [
            'field_date', 'field_datetime', 'field_int', 'field_str', ]
        site_reference_configs.registry = {}
        reference = ReferenceModelConfig(fields=fields, name=name)
        site_reference_configs.register(reference=reference)
        self.assertEqual(fields, site_reference_configs.get_fields(name=name))

    def test_site_reregister(self):
        name = 'edc_reference.crfone'
        site_reference_configs.registry = {}
        reference = ReferenceModelConfig(
            name=name,
            fields=['field_date', 'field_datetime',
                    'field_int', 'field_str', ])
        site_reference_configs.register(reference=reference)
        self.assertRaises(
            AlreadyRegistered, site_reference_configs.register, reference=reference)

    def test_autodiscover(self):
        site_reference_configs.registry = {}
        site_reference_configs.loaded = False
        site_reference_configs.autodiscover('tests.reference_model_configs')
        self.assertTrue(site_reference_configs.loaded)
        self.assertIn('edc_reference.erik', list(
            site_reference_configs.registry))

        self.assertEqual(['f1', 'f2', 'f3', 'f4'],
                         site_reference_configs.get_fields('edc_reference.erik'))

    def test_autodiscover_bad(self):
        self.assertRaises(
            TypeError,
            site_reference_configs.autodiscover,
            module_name='tests.reference_model_configs_bad')

    def test_check_ok(self):
        name = 'edc_reference.crfone'
        fields = [
            'field_str', 'field_date', 'field_datetime', 'field_int']
        site_reference_configs.registry = {}
        reference = ReferenceModelConfig(fields=fields, name=name)
        try:
            reference.check()
        except (ReferenceFieldValidationError, ReferenceModelValidationError) as e:
            self.fail(
                f'Reference validation error unexpectedly raised. Got{e}')

    def test_check_ok2(self):
        name = 'edc_reference.crfone'
        site_reference_configs.registry = {}
        reference = ReferenceModelConfig(
            fields=['field_str', 'field_date', 'field_datetime', 'field_int'],
            name=name)
        site_reference_configs.register(reference=reference)
        site_reference_configs.check()

    def test_check_bad_model(self):
        name = 'edc_reference.erik'
        fields = ['f1']
        reference = ReferenceModelConfig(fields=fields, name=name)
        self.assertRaises(
            ReferenceModelValidationError,
            reference.check)

    def test_check_bad_fields(self):
        name = 'edc_reference.crfone'
        fields = ['f1']
        site_reference_configs.registry = {}
        reference = ReferenceModelConfig(fields=fields, name=name)
        self.assertRaises(
            ReferenceFieldValidationError,
            reference.check)

    def test_raises_on_duplicate_field_name(self):
        name = 'edc_reference.crfone'
        fields = ['f1', 'f1']
        self.assertRaises(
            ReferenceDuplicateField,
            ReferenceModelConfig, fields=fields, name=name)

    def test_not_registered_for_fields(self):
        name = 'edc_reference.crfone'
        site_reference_configs.registry = {}
        self.assertRaises(
            SiteReferenceConfigError,
            site_reference_configs.get_fields, name=name)

    def test_not_registered_for_reference_model(self):
        name = 'edc_reference.crfone'
        site_reference_configs.registry = {}
        self.assertRaises(
            SiteReferenceConfigError,
            site_reference_configs.get_reference_model, name=name)

    def test_not_registered_for_reregister(self):
        name = 'edc_reference.crfone'
        fields = ['f1']
        reference = ReferenceModelConfig(fields=fields, name=name)
        site_reference_configs.registry = {}
        self.assertRaises(
            ReferenceConfigNotRegistered,
            site_reference_configs.reregister, reference)

    def test_get_config(self):
        site_reference_configs.registry = {}
        name = 'edc_reference.crfone'
        fields = ['f1']
        reference = ReferenceModelConfig(fields=fields, name=name)
        site_reference_configs.register(reference)
        self.assertEqual(
            reference, site_reference_configs.get_config(name='edc_reference.crfone'))
        self.assertRaises(
            SiteReferenceConfigError,
            site_reference_configs.get_config, name=None)

    def test_register_reference_models_from_visit_schedule(self):
        site_reference_configs.registry = {}
        site_reference_configs.register_from_visit_schedule(
            visit_models={'edc_appointment.appointment': 'edc_reference.subjectvisit'})
        self.assertTrue(
            site_reference_configs.get_config(name='edc_reference.crfone'))

    def test_reregister_reference_models_from_visit_schedule(self):
        site_reference_configs.registry = {}
        site_reference_configs.register_from_visit_schedule(
            visit_models={'edc_appointment.appointment': 'edc_reference.subjectvisit'})
        # do again
        site_reference_configs.register_from_visit_schedule(
            visit_models={'edc_appointment.appointment': 'edc_reference.subjectvisit'})
        self.assertTrue(
            site_reference_configs.get_config(name='edc_reference.crfone'))

    def test_add_fields_to_reference_config(self):
        name = 'edc_reference.crfone'
        fields = ['f1']
        site_reference_configs.registry = {}
        reference_config = ReferenceModelConfig(fields=fields, name=name)
        site_reference_configs.register(reference_config)
        self.assertEqual(reference_config.field_names, ['f1'])
        site_reference_configs.add_fields_to_config(
            name=name, fields=['f2', 'f3'])
        self.assertEqual(reference_config.field_names, ['f1', 'f2', 'f3'])



from .fieldset import Fieldset, FieldsetError
from .longitudinal_refset import LongitudinalRefset
from .longitudinal_refset import NoRefsetObjectsExist, InvalidOrdering
from .refset import Refset, RefsetError, RefsetOverlappingField



# from django.test import TestCase, tag
#
# from edc_base.utils import get_utcnow
# from edc_sync.models import OutgoingTransaction
#
# from .models import SubjectVisit
#
# from ..reference_model_config import ReferenceModelConfig
# from ..site import site_reference_configs
# from .models import TestModel
#
#
# class TestSyncModels(TestCase):
#
#     def setUp(self):
#         site_reference_configs.registry = {}
#         self.subject_identifier = '1'
#         subjectvisit_reference = ReferenceModelConfig(
#             name='edc_reference.subjectvisit',
#             fields=['report_datetime', 'visit_code'])
#         self.testmodel_reference = ReferenceModelConfig(
#             name='edc_reference.testmodel', fields=['field_str', 'report_datetime'])
#         self.crfone_reference = ReferenceModelConfig(
#             name='edc_reference.crfone',
#             fields=['field_str', 'field_date', 'field_datetime',
#                    'field_int', 'report_datetime'])
#         site_reference_configs.register(subjectvisit_reference)
#         site_reference_configs.register(self.testmodel_reference)
#         site_reference_configs.register(self.crfone_reference)
#         self.subject_visit = SubjectVisit.objects.create(
#             subject_identifier=self.subject_identifier,
#             visit_code='code')
#
#     def test_reference_synchronizes(self):
#         self.assertEqual(OutgoingTransaction.objects.filter(
#             tx_name='edc_reference.reference').count(), 6)
#         TestModel.objects.create(
#             subject_visit=self.subject_visit,
#             field_str='erik')
#         self.assertEqual(OutgoingTransaction.objects.filter(
#             tx_name='edc_reference.reference').count(), 12)



from django.core.checks import Warning

from .site import site_reference_configs


def check_site_reference_configs(app_configs, **kwargs):
    errors = []
    site_results = site_reference_configs.check()
    for result in site_results.values():
        errors.append(
            Warning(result, id=f'edc_reference.001'))
    return errors



from uuid import uuid4

from datetime import date
from dateutil.relativedelta import relativedelta
from django.core.exceptions import ObjectDoesNotExist
from django.test import TestCase, tag
from edc_base.utils import get_utcnow
from edc_lab.models.panel import Panel

from ..models import Reference, ReferenceFieldDatatypeNotFound
from ..reference import ReferenceDeleter, ReferenceGetter
from ..reference import ReferenceObjectDoesNotExist
from ..reference import ReferenceUpdater, ReferenceFieldNotFound
from ..reference_model_config import ReferenceDuplicateField, ReferenceFieldValidationError
from ..reference_model_config import ReferenceModelConfig
from ..site import site_reference_configs
from .models import CrfOne, SubjectVisit
from .models import CrfWithUnknownDatatype, TestModel, SubjectRequisition


class TestReferenceModel(TestCase):

    def setUp(self):
        self.panel_cd4 = Panel.objects.create(name='cd4')
        self.panel_vl = Panel.objects.create(name='vl')
        self.panel_wb = Panel.objects.create(name='wb')
        site_reference_configs.registry = {}
        self.subject_identifier = '1'
        subjectvisit_reference = ReferenceModelConfig(
            name='edc_reference.subjectvisit',
            fields=['report_datetime', 'visit_code'])
        self.testmodel_reference = ReferenceModelConfig(
            name='edc_reference.testmodel', fields=[
                'field_str', 'report_datetime'])
        self.crfone_reference = ReferenceModelConfig(
            name='edc_reference.crfone',
            fields=['field_str', 'field_date', 'field_datetime',
                    'field_int', 'report_datetime'])
        site_reference_configs.register(subjectvisit_reference)
        site_reference_configs.register(self.testmodel_reference)
        site_reference_configs.register(self.crfone_reference)
        self.subject_visit = SubjectVisit.objects.create(
            subject_identifier=self.subject_identifier,
            visit_code='code')

    def test_updater_repr(self):
        model_obj = TestModel.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        self.assertTrue(
            repr(ReferenceUpdater(model_obj=model_obj)))

    def test_model_repr(self):
        model_obj = TestModel.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        ReferenceUpdater(model_obj=model_obj)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            field_name='field_str')
        self.assertTrue(repr(reference))

    def test_updater_creates_reference(self):
        model_obj = TestModel.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        ReferenceUpdater(model_obj=model_obj)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            field_name='field_str')
        self.assertEqual(reference.value, 'erik')

    def test_updater_updates_reference(self):
        model_obj = TestModel.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        ReferenceUpdater(model_obj=model_obj)
        model_obj.field_str = 'bob'
        ReferenceUpdater(model_obj=model_obj)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            field_name='field_str')
        self.assertEqual(reference.value, 'bob')

    def test_updater_with_bad_field_name(self):
        site_reference_configs.registry = {}
        self.testmodel_reference = ReferenceModelConfig(
            name='edc_reference.testmodel', fields=['blah'])
        site_reference_configs.register(self.testmodel_reference)
        self.assertRaises(
            ReferenceFieldNotFound,
            TestModel.objects.create,
            subject_visit=self.subject_visit,
            field_str='erik')

    def test_deleter(self):
        model_obj = TestModel.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        ReferenceUpdater(model_obj=model_obj)
        model_obj.delete()
        ReferenceDeleter(model_obj=model_obj)
        try:
            reference = Reference.objects.get(
                identifier=self.subject_identifier,
                timepoint=self.subject_visit.visit_code,
                field_name='field_str')
        except ObjectDoesNotExist:
            pass
        else:
            self.fail(f'Object unexpectedly exists. Got {reference}')

    def test_model_mixin_deleter(self):
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        self.assertGreater(Reference.objects.filter(
            model='edc_reference.crfone').count(), 2)
        crf_one.delete()
        self.assertEqual(0, Reference.objects.filter(
            model='edc_reference.crfone').count())

    def test_model_creates_reference(self):
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        self.assertEqual(
            len(site_reference_configs.get_fields(
                'edc_reference.crfone')), 5)
        self.assertEqual(Reference.objects.all().count(), 7)

    def test_model_creates_reference2(self):
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        try:
            reference = Reference.objects.get(
                identifier=self.subject_identifier,
                timepoint=self.subject_visit.visit_code,
                report_datetime=self.subject_visit.report_datetime,
                model='edc_reference.crfone',
                field_name='field_str')
        except ObjectDoesNotExist as e:
            self.fail(f'ObjectDoesNotExist unexpectedly raised. Got {e}')
        self.assertEqual(reference.value, 'erik')

    def test_model_get_for_report_datetime(self):
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        try:
            reference = Reference.objects.get(
                identifier=self.subject_identifier,
                timepoint=self.subject_visit.visit_code,
                report_datetime=self.subject_visit.report_datetime,
                model='edc_reference.crfone',
                field_name='report_datetime')
        except ObjectDoesNotExist as e:
            self.fail(f'ObjectDoesNotExist unexpectedly raised. Got {e}')
        self.assertEqual(reference.value, self.subject_visit.report_datetime)

    def test_model_updates_reference(self):
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str='erik')
        crf_one.field_str = 'bob'
        crf_one.save()
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            report_datetime=self.subject_visit.report_datetime,
            field_name='field_str')
        self.assertEqual(reference.value, 'bob')

    def test_model_creates_for_date(self):
        dte = date.today()
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_date=dte)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            report_datetime=self.subject_visit.report_datetime,
            field_name='field_date')
        self.assertEqual(reference.value, dte)

    def test_model_creates_for_datetime(self):
        dte = get_utcnow()
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_datetime=dte)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            report_datetime=self.subject_visit.report_datetime,
            field_name='field_datetime')
        self.assertEqual(reference.value, dte)

    def test_model_creates_for_int(self):
        integer = 100
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=integer)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            report_datetime=self.subject_visit.report_datetime,
            field_name='field_int')
        self.assertEqual(reference.value, integer)

    def test_model_creates_for_all(self):
        strval = 'erik'
        integer = 100
        dte = date.today()
        dtetime = get_utcnow()
        field_datetime = dtetime - relativedelta(years=1)
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            report_datetime=dtetime - relativedelta(years=10),
            field_str=strval,
            field_int=integer,
            field_date=dte,
            field_datetime=field_datetime)
        for field_name in site_reference_configs.get_fields('edc_reference.crfone'):
            reference = Reference.objects.get(
                identifier=self.subject_identifier,
                model='edc_reference.crfone',
                timepoint=self.subject_visit.visit_code,
                report_datetime=self.subject_visit.report_datetime,
                field_name=field_name)
            self.assertIn(reference.value, [
                strval, integer, dte, field_datetime,
                self.subject_visit.report_datetime],
                msg=f'field_name={field_name}')

    def test_model_creates_and_gets_for_report_datetime(self):
        """Assert uses subject visit report_datetime and not
        CRF report_datetime for field_name='report_datetime'.
        """
        dtetime = get_utcnow()
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            report_datetime=get_utcnow() - relativedelta(years=10),
            field_datetime=dtetime)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            model='edc_reference.crfone',
            timepoint=self.subject_visit.visit_code,
            report_datetime=self.subject_visit.report_datetime,
            field_name='report_datetime')
        self.assertEqual(reference.value, self.subject_visit.report_datetime)

        getter = ReferenceGetter(
            field_name='report_datetime',
            model_obj=crf_one)
        self.assertEqual(getter.value, self.subject_visit.report_datetime)

    def test_model_create_handles_none(self):
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=None)
        reference = Reference.objects.get(
            identifier=self.subject_identifier,
            timepoint=self.subject_visit.visit_code,
            report_datetime=self.subject_visit.report_datetime,
            field_name='field_int')
        self.assertEqual(reference.value, None)

    def test_model_raises_on_duplicate_field_name(self):
        self.assertRaises(
            ReferenceDuplicateField,
            ReferenceModelConfig,
            name='edc_reference.crfwithduplicatefield',
            fields=['field_int', 'field_int', 'field_datetime', 'field_str'])

    def test_model_raises_on_bad_field_name(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.crfwithbadfield',
            fields=['blah1', 'blah2', 'blah3', 'blah4'])
        self.assertRaises(
            ReferenceFieldValidationError,
            reference_config.check)

    def test_model_raises_on_bad_field_name_checked_by_site(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.crfwithbadfield',
            fields=['blah1', 'blah2', 'blah3', 'blah4'])
        site_reference_configs.register(reference_config)
        errors = site_reference_configs.check()
        self.assertIsNotNone(errors)

    def test_raises_on_missing_model_mixin(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.TestModelBad',
            fields=['report_datetime'])
        site_reference_configs.register(reference_config)
        errors = site_reference_configs.check()
        self.assertIsNotNone(errors)

    def test_model_raises_on_unknown_field_datatype(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.CrfWithUnknownDatatype',
            fields=['field_decimal'])
        site_reference_configs.register(reference_config)
        self.assertRaises(
            ReferenceFieldDatatypeNotFound,
            CrfWithUnknownDatatype.objects.create,
            subject_visit=self.subject_visit,
            field_decimal=3.2)

    def test_getter_repr(self):
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=100)
        reference = ReferenceGetter(
            field_name='field_int',
            model_obj=crf_one)
        self.assertTrue(repr(reference))

    def test_report_datetime_uses_visit_report_datetime(self):
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=100)
        report_datetimes = []
        for obj in Reference.objects.filter(model__icontains='crfone'):
            report_datetimes.append(obj.report_datetime)
        self.assertGreater(len(report_datetimes), 0)
        for report_datetime in report_datetimes:
            self.assertEqual(
                report_datetime, self.subject_visit.report_datetime)

    def test_reference_getter_sets_attr(self):
        integer = 100
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=integer)
        reference = ReferenceGetter(
            field_name='field_int',
            model_obj=crf_one)
        self.assertEqual(reference.field_int, integer)

    def test_reference_getter_sets_attr_even_if_none(self):
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit)
        reference = ReferenceGetter(
            field_name='field_int',
            model_obj=crf_one)
        self.assertEqual(reference.field_int, None)

    def test_site_checks_no_fields_raises(self):
        name = 'edc_reference.crfone'
        site_reference_configs.registry = {}
        self.assertRaises(
            ReferenceFieldValidationError,
            ReferenceModelConfig,
            fields=[],
            name=name)

    def test_site_checks_no_fields_raises2(self):
        name = 'edc_reference.crfone'
        site_reference_configs.registry = {}
        self.assertRaises(
            ReferenceFieldValidationError,
            ReferenceModelConfig,
            name=name)

    def test_reference_getter_without_crf_type_model(self):
        integer = 100
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=integer)
        reference = ReferenceGetter(
            name='edc_reference.crfone',
            field_name='field_int',
            visit_obj=crf_one.visit)
        self.assertEqual(reference.field_int, integer)

    def test_reference_getter_without_using_model_obj(self):
        integer = 100
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=integer)
        reference = ReferenceGetter(
            name='edc_reference.crfone',
            field_name='field_int',
            subject_identifier=self.subject_identifier,
            report_datetime=crf_one.visit.report_datetime,
            visit_code=crf_one.visit.visit_code)
        self.assertEqual(reference.field_int, integer)

    def test_reference_getter_with_bad_field_raises(self):
        integer = 100
        crf_one = CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_int=integer)
        self.assertRaises(
            ReferenceObjectDoesNotExist,
            reference=ReferenceGetter,
            name='edc_reference.crfone',
            field_name='blah',
            model_obj=crf_one)
        try:
            ReferenceGetter(
                field_name='blah',
                name='edc_reference.crfone',
                model_obj=crf_one)
        except ReferenceObjectDoesNotExist:
            pass

    def test_reference_getter_doesnotexist(self):
        self.assertRaises(
            ReferenceObjectDoesNotExist,
            reference=ReferenceGetter,
            field_name='blah',
            model='edc_reference.crfone',
            model_obj=self.subject_visit)

    def test_model_manager_crf(self):
        strval = 'erik'
        integer = 100
        dte = date.today()
        dtetime = get_utcnow()
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str=strval,
            field_int=integer,
            field_date=dte,
            field_datetime=dtetime)
        qs = Reference.objects.filter_crf_for_visit(
            'edc_reference.crfone', self.subject_visit)
        self.assertEqual(qs.count(), 5)

    def test_model_manager_crf_by_field(self):
        strval = 'erik'
        integer = 100
        dte = date.today()
        dtetime = get_utcnow()
        CrfOne.objects.create(
            subject_visit=self.subject_visit,
            field_str=strval,
            field_int=integer,
            field_date=dte,
            field_datetime=dtetime)
        obj = Reference.objects.get_crf_for_visit(
            'edc_reference.crfone', self.subject_visit, 'field_str')
        self.assertEqual(obj.value, 'erik')
        obj = Reference.objects.get_crf_for_visit(
            'edc_reference.crfone', self.subject_visit, 'field_int')
        self.assertEqual(obj.value, integer)
        obj = Reference.objects.get_crf_for_visit(
            'edc_reference.crfone', self.subject_visit, 'field_date')
        self.assertEqual(obj.value, dte)
        obj = Reference.objects.get_crf_for_visit(
            'edc_reference.crfone', self.subject_visit, 'field_datetime')
        self.assertEqual(obj.value, dtetime)
        obj = Reference.objects.get_crf_for_visit(
            'edc_reference.crfone', self.subject_visit, 'blah')
        self.assertIsNone(obj)

    def test_model_manager_requisition(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.subjectrequisition.cd4',
            fields=['panel'])
        site_reference_configs.register(reference_config)
        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel_cd4)
        obj = Reference.objects.get_requisition_for_visit(
            'edc_reference.subjectrequisition.cd4', self.subject_visit)
        self.assertEqual(obj.value, self.panel_cd4.id)
        obj = Reference.objects.get_requisition_for_visit(
            'edc_reference.subjectrequisition.blah', self.subject_visit)
        self.assertIsNone(obj)

    def test_requisition_creates_two(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.subjectrequisition.cd4',
            fields=['panel'])
        site_reference_configs.register(reference_config)
        reference_config = ReferenceModelConfig(
            name='edc_reference.subjectrequisition.vl',
            fields=['panel'])
        site_reference_configs.register(reference_config)
        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel_cd4)
        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel_vl)
        obj = Reference.objects.get_requisition_for_visit(
            'edc_reference.subjectrequisition.vl', self.subject_visit)
        self.assertEqual(obj.value, self.panel_vl.id)
        obj = Reference.objects.get_requisition_for_visit(
            'edc_reference.subjectrequisition.cd4', self.subject_visit)
        self.assertEqual(obj.value, self.panel_cd4.id)
        obj = Reference.objects.get_requisition_for_visit(
            'edc_reference.subjectrequisition.blah', self.subject_visit)
        self.assertIsNone(obj)

    def test_requisition_creates2(self):
        reference_config = ReferenceModelConfig(
            name='edc_reference.subjectrequisition.cd4',
            fields=['panel'])
        site_reference_configs.register(reference_config)
        reference_config = ReferenceModelConfig(
            name='edc_reference.subjectrequisition.vl',
            fields=['panel'])
        site_reference_configs.register(reference_config)
        reference_config = ReferenceModelConfig(
            name='edc_reference.subjectrequisition.wb',
            fields=['panel'])
        site_reference_configs.register(reference_config)

        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel_vl)
        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel_cd4)
        SubjectRequisition.objects.create(
            subject_visit=self.subject_visit,
            panel=self.panel_wb)

        for panel in [self.panel_cd4, self.panel_vl, self.panel_wb]:
            with self.subTest(panel=panel):
                Reference.objects.get(
                    model=f'edc_reference.subjectrequisition.{panel.name}',
                    report_datetime=self.subject_visit.report_datetime,
                    timepoint=self.subject_visit.visit_code,
                    field_name='panel',
                    value_uuid=panel.id)



from edc_navbar import Navbar, NavbarItem, site_navbars


reference = Navbar(name='edc_reference')

reference.append_item(
    NavbarItem(name='reference',
               title='reference',
               label='reference',
               url_name='edc_reference:home_url'))

site_navbars.register(reference)



from django.contrib.admin.sites import AdminSite


class EdcReferenceAdminSite(AdminSite):
    site_header = 'Edc Reference'
    site_title = 'Edc Reference'
    index_title = 'Edc Reference'
    site_url = '/administration/'


edc_reference_admin = EdcReferenceAdminSite(name='edc_reference_admin')



import copy
import sys

from django.apps import apps as django_apps
from django.utils.module_loading import import_module, module_has_submodule
from edc_visit_schedule.site_visit_schedules import site_visit_schedules
from edc_lab.site_labs import site_labs

from .reference_model_config import ReferenceDuplicateField, ReferenceModelValidationError
from .reference_model_config import ReferenceFieldValidationError
from .reference_model_config import ReferenceModelConfig
from .reference_model_config import ReferenceFieldAlreadyAdded


class AlreadyRegistered(Exception):
    pass


class ReferenceConfigNotRegistered(Exception):
    pass


class RegistryNotLoaded(Exception):
    pass


class SiteReferenceConfigImportError(Exception):
    pass


class SiteReferenceConfigError(Exception):
    pass


class ReferenceUpdater:

    def update(self, name=None, fields=None, get_config=None):
        try:
            reference = get_config(name=name)
        except SiteReferenceConfigError:
            reference = ReferenceModelConfig(name=name, fields=fields)
        else:
            for field in fields:
                try:
                    reference.add_fields([field])
                except ReferenceFieldAlreadyAdded:
                    pass
        return reference


class Site:

    reference_updater = ReferenceUpdater()

    def __init__(self):
        self.registry = {}
        self.loaded = False
        self.registered_from_visit_schedules = False

    def register(self, reference=None):
        if reference.name in self.registry:
            raise AlreadyRegistered(
                f'Reference fields have already been registered. '
                f'Got {reference.name}')
        self.registry.update({reference.name: reference})
        self.loaded = True

    def reregister(self, reference=None):
        if reference.name not in self.registry:
            raise ReferenceConfigNotRegistered(
                f'Reregister failed. Reference model configuration has not '
                f'been registered. Got {reference.name}')
        self.registry.update({reference.name: reference})

    def unregister(self, name=None):
        if name not in self.registry:
            raise ReferenceConfigNotRegistered(
                f'Unregister failed. Reference model configuration has not '
                f'been registered. Got {name}')
        self.registry = {k: v for k,
                         v in self.registry.items() if k != name}

    def get_config(self, name=None):
        try:
            reference_config = self.registry.get(name)
        except AttributeError:
            reference_config = None
        if not reference_config:
            raise SiteReferenceConfigError(
                f'Model not registered. Got {name}. Expected one of '
                f'{list(self.registry.keys())}.')
        return reference_config

    def get_fields(self, name=None):
        """Returns a list of fields associated with the
        reference configuration of "model".
        """
        return self.get_config(name=name).field_names

    def get_reference_model(self, name=None):
        """Returns the reference model associated with the
        reference configuration of "model".
        """
        return self.get_config(name=name).reference_model

    def check(self):
        """Validates the reference data for all classes in the
        registry.
        """
        errors = {}
        for name in self.registry:
            reference_config = self.get_config(name)
            try:
                reference_config.check()
            except (ReferenceDuplicateField, ReferenceModelValidationError,
                    ReferenceFieldValidationError) as e:
                errors.update({name: str(e)})
        return errors

    def autodiscover(self, module_name=None):
        """Autodiscovers classes in the reference_model_configs.py file of any
        INSTALLED_APP.
        """
        module_name = module_name or 'reference_model_configs'
        sys.stdout.write(f' * checking for {module_name} ...\n')
        for app in django_apps.app_configs:
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(
                        site_reference_configs.registry)
                    import_module(f'{app}.{module_name}')
                    sys.stdout.write(
                        f' * registered reference model configs from application \'{app}\'\n')
                except Exception as e:
                    if f'No module named \'{app}.{module_name}\'' not in str(e):
                        site_reference_configs.registry = before_import_registry
                        if module_has_submodule(mod, module_name):
                            raise
            except ModuleNotFoundError:
                pass

    def register_from_visit_schedule(self, visit_models=None, extra_requisition_fields=None):
        """Registers CRFs and Requisitions for all visits
        under schedules using this visit model.

        Note: Unscheduled and PRN forms are automatically add as well.
        """
        requisition_fields = ['requisition_datetime', 'panel', 'is_drawn',
                              'reason_not_drawn']
        requisition_fields.extend(extra_requisition_fields or [])
        requisition_fields = list(set(requisition_fields))

        self.registered_visit_model = True
        site_labs.autodiscover(verbose=False)
        site_visit_schedules.autodiscover(verbose=False)
        for visit_schedule in site_visit_schedules.registry.values():
            for schedule in visit_schedule.schedules.values():
                reference = self.reference_updater.update(
                    name=visit_models[schedule.appointment_model],
                    fields=['report_datetime'],
                    get_config=self.get_config)
                self._register_if_new(reference)
                for visit in schedule.visits.values():
                    for crf in visit.all_crfs:
                        reference = self.reference_updater.update(
                            name=crf.model,
                            fields=['report_datetime'],
                            get_config=self.get_config)
                        self._register_if_new(reference)
                    for requisition in visit.all_requisitions:
                        if not requisition.panel.requisition_model:
                            raise SiteReferenceConfigError(
                                'Requisition\'s panel \'requisition_model\' attribute '
                                f'not set. See "{requisition}". Has the requisition '
                                'been added to a lab profile and registered? Is the '
                                'APP in INSTALLED_APPS? Currently '
                                f'registered lab profiles are {list(site_labs._registry)}.')
                        reference = self.reference_updater.update(
                            name=f'{requisition.model}.{requisition.panel.name}',
                            fields=requisition_fields,
                            get_config=self.get_config)
                        self._register_if_new(reference)

    def _register_if_new(self, reference):
        try:
            self.register(reference)
        except AlreadyRegistered:
            self.reregister(reference)

    def add_fields_to_config(self, name=None, fields=None):
        reference_config = self.get_config(name=name)
        reference_config.add_fields(fields)


site_reference_configs = Site()



from datetime import date, datetime
from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from ..site import site_reference_configs


class ReferenceTestHelperError(Exception):
    pass


class ReferenceTestHelper:

    visit_model = None

    field_types = {
        'CharField': (str),
        'DateTimeField': (datetime),
        'DateField': (date),
        'IntegerField': (int)}

    def __init__(self, visit_model=None, subject_identifier=None):
        if visit_model:
            self.visit_model = visit_model
        self.subject_identifier = subject_identifier
        self.app_label = visit_model.split('.')[0]
        self.reference_model_cls = django_apps.get_model(
            site_reference_configs.get_reference_model(name=visit_model))

    def get(self, **kwargs):
        return self.reference_model_cls.objects.get(**kwargs)

    def filter(self, **kwargs):
        return self.reference_model_cls.objects.filter(**kwargs)

    def create_for_model(self, reference_name=None, report_datetime=None,
                         visit_code=None, **options):
        for field_name in site_reference_configs.get_fields(
                name=reference_name):
            reference = self.reference_model_cls.objects.create(
                model=reference_name,
                identifier=self.subject_identifier,
                report_datetime=report_datetime,
                timepoint=visit_code,
                field_name=field_name)
            if field_name != 'report_datetime':
                try:
                    value, internal_type = options.get(field_name)
                    if internal_type not in self.field_types:
                        raise TypeError(
                            f'Invalid internal type. Got \'{internal_type}\'')
                except (TypeError, ValueError) as e:
                    value = options.get(field_name)
                    internal_type = None
                    if value:
                        for field_type, datatypes in self.field_types.items():
                            if isinstance(value, datatypes):
                                internal_type = field_type
                                break
                        if not internal_type:
                            raise TypeError(
                                f'{e}. Got field_name={field_name}, value={value}')
                if value:
                    reference.update_value(
                        value=value, internal_type=internal_type)
        return self.reference_model_cls.objects.filter(
            model=reference_name, identifier=self.subject_identifier,
            report_datetime=report_datetime)

    def update_for_model(self, reference_name=None,
                         report_datetime=None, visit_code=None,
                         valueset=None):
        for field_name, internal_type, value in valueset:
            try:
                reference = self.reference_model_cls.objects.get(
                    model=reference_name,
                    identifier=self.subject_identifier,
                    report_datetime=report_datetime,
                    timepoint=visit_code,
                    field_name=field_name)
            except ObjectDoesNotExist:
                reference = self.reference_model_cls.objects.create(
                    model=reference_name,
                    identifier=self.subject_identifier,
                    report_datetime=report_datetime,
                    timepoint=visit_code,
                    field_name=field_name)
            reference.update_value(value=value, internal_type=internal_type)

    def create_visit(self, report_datetime=None, timepoint=None):
        reference = self.reference_model_cls.objects.create(
            identifier=self.subject_identifier,
            model=self.visit_model,
            report_datetime=report_datetime,
            timepoint=timepoint,
            field_name='report_datetime')
        return reference



from django.db import models

from .reference import ReferenceDeleter, ReferenceUpdater


class ReferenceModelMixinError(Exception):
    pass


class ReferenceModelMixin(models.Model):

    reference_deleter_cls = ReferenceDeleter
    reference_updater_cls = ReferenceUpdater

    def save(self, *args, **kwargs):
        self.model_reference_validate()
        self.reference_updater_cls(model_obj=self)
        super().save(*args, **kwargs)

    @property
    def reference_name(self):
        return self._meta.label_lower

    def model_reference_validate(self):
        if 'panel' in [f.name for f in self._meta.get_fields()]:
            raise ReferenceModelMixinError(
                'Detected field panel. Is this a requisition?. '
                'Use RequisitionReferenceModelMixin '
                'instead of ReferenceModelMixin')

    class Meta:
        abstract = True


class RequisitionReferenceModelMixin(ReferenceModelMixin, models.Model):

    @property
    def reference_name(self):
        return f'{self._meta.label_lower}.{self.panel.name}'

    def model_reference_validate(self):
        if 'panel' not in [f.name for f in self._meta.get_fields()]:
            raise ReferenceModelMixinError(
                'Did not detect field panel. Is this a CRF?. '
                'Use ReferenceModelMixin '
                'instead of RequisitionReferenceModelMixin')

    class Meta:
        abstract = True



from ..site import site_reference_configs
from .reference_getter import ReferenceGetter


class ReferenceFieldNotFound(Exception):
    pass


class ReferenceUpdaterModelError(Exception):
    pass


class ReferenceUpdater:
    """Updates or creates each reference model instance; one for
    each field in `edc_reference` for this model_obj.
    """

    getter_cls = ReferenceGetter
    crf_visit_attr = 'visit'

    def __init__(self, model_obj=None):

        reference_fields = site_reference_configs.get_fields(
            name=model_obj.reference_name)
        # loop through fields and update or create each
        # reference model instance
        for field_name in reference_fields:
            try:
                field_obj = [fld for fld in model_obj._meta.get_fields()
                             if fld.name == field_name][0]
            except IndexError:
                raise ReferenceFieldNotFound(
                    f'Reference field not found on model. Got \'{field_name}\'. '
                    f'See reference {model_obj.reference_name}. '
                    f'Model fields are {[fld.name for fld in model_obj._meta.get_fields()]}')
            reference = self.getter_cls(
                model_obj=model_obj,
                field_name=field_name,
                create=True)
            if field_obj.name == 'report_datetime':
                try:
                    visit = getattr(model_obj, self.crf_visit_attr)
                    value = getattr(visit, field_name)
                except AttributeError:
                    value = getattr(model_obj, field_name)
            else:
                value = getattr(model_obj, field_name)
            try:
                value = value.pk
            except AttributeError:
                internal_type = field_obj.get_internal_type()
                related_name = None
            else:
                internal_type = 'UUIDField'
                related_name = getattr(model_obj, field_name)._meta.label_lower
            reference.object.update_value(
                internal_type=internal_type,
                value=value,
                related_name=related_name)
            reference.object.save()



from django.apps import apps as django_apps
from django.db import transaction

from ..site import site_reference_configs


class ReferenceDeleter:

    """A class to delete all instances for the reference
    model for this model instance.

    See signals.
    """

    def __init__(self, model_obj=None):
        reference_model = site_reference_configs.get_reference_model(
            name=model_obj.reference_name)
        self.reference_model_cls = django_apps.get_model(reference_model)
        self.model_obj = model_obj
        self.reference_objects = self.reference_model_cls.objects.filter(
            **self.options)
        with transaction.atomic():
            self.reference_objects.delete()

    @property
    def options(self):
        return dict(
            identifier=self.model_obj.visit.subject_identifier,
            report_datetime=self.model_obj.visit.report_datetime,
            timepoint=self.model_obj.visit.visit_code,
            model=self.model_obj.reference_name)



from .refsets import LongitudinalRefset, NoRefsetObjectsExist
from .reference import ReferenceDeleter
from .reference import ReferenceGetter
from .reference import ReferenceUpdater
from .reference import ReferenceFieldNotFound
from .reference_model_config import ReferenceModelConfig
from .site import site_reference_configs, ReferenceDuplicateField



from django.core.management.base import BaseCommand

from edc_reference.populater import Populater
from edc_constants.constants import YES, NO


class Command(BaseCommand):

    help = 'Populates the reference model'

    def add_arguments(self, parser):

        parser.add_argument(
            '--names',
            dest='names',
            nargs='*',
            default=None,
            help=(
                'run for a select list of reference names (label_lower or panel_name)'),
        )
        parser.add_argument(
            '--exclude_names',
            dest='exclude_names',
            nargs='*',
            default=None,
            help='exclude reference names (label_lower or panel_name)')

        parser.add_argument(
            '--skip-existing',
            dest='skip_existing',
            nargs='?',
            choices=[YES, NO],
            default=NO,
            const=YES,
            help=(
                f'skip reference name if records already exist (Default: {NO})'),
        )

        parser.add_argument(
            '--summarize',
            dest='summarize',
            nargs='?',
            choices=[YES, NO],
            const=YES,
            default=NO,
            help=(f'Summarize existing data (Default: {NO})'),
        )

        parser.add_argument(
            '--dry-run',
            dest='dry_run',
            nargs='?',
            choices=[YES, NO],
            const=YES,
            default=NO,
            help=(f'Summarize existing data (Default: {NO})'),
        )

        parser.add_argument(
            '--delete-existing',
            dest='delete_existing',
            nargs='?',
            choices=[YES, NO],
            const=YES,
            default=NO,
            help=(f'Delete existing data (Default: {NO})'),
        )

    def handle(self, *args, **options):
        names = options.get('names')
        exclude_names = options.get('exclude_names')
        skip_existing = options.get('skip_existing')
        skip_existing = None if skip_existing == NO else YES
        delete_existing = options.get('delete_existing')
        delete_existing = None if delete_existing == NO else YES
        summarize = options.get('summarize')
        summarize = None if summarize == NO else YES
        dry_run = options.get('dry_run')
        dry_run = None if dry_run == NO else YES
        populater = Populater(
            names=names,
            exclude_names=exclude_names,
            skip_existing=skip_existing,
            delete_existing=delete_existing,
            dry_run=dry_run)
        if summarize:
            populater.summarize()
        else:
            populater.populate()



# Generated by Django 2.0 on 2017-12-13 20:34

import _socket
from django.db import migrations, models
import django_revision.revision_field
import edc_model_fields.fields.hostname_modification_field
import edc_model_fields.fields.userfield
import edc_model_fields.fields.uuid_auto_field
import edc_base.utils


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Reference',
            fields=[
                ('created', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('modified', models.DateTimeField(blank=True, default=edc_base.utils.get_utcnow)),
                ('user_created', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user created')),
                ('user_modified', edc_model_fields.fields.userfield.UserField(blank=True, help_text='Updated by admin.save_model', max_length=50, verbose_name='user modified')),
                ('hostname_created', models.CharField(blank=True, default=_socket.gethostname, help_text='System field. (modified on create only)', max_length=60)),
                ('hostname_modified', edc_model_fields.fields.hostname_modification_field.HostnameModificationField(blank=True, help_text='System field. (modified on every save)', max_length=50)),
                ('revision', django_revision.revision_field.RevisionField(blank=True, editable=False, help_text='System field. Git repository tag:branch:commit.', max_length=75, null=True, verbose_name='Revision')),
                ('device_created', models.CharField(blank=True, max_length=10)),
                ('device_modified', models.CharField(blank=True, max_length=10)),
                ('id', edc_model_fields.fields.uuid_auto_field.UUIDAutoField(blank=True, editable=False, help_text='System auto field. UUID primary key.', primary_key=True, serialize=False)),
                ('identifier', models.CharField(max_length=50)),
                ('timepoint', models.CharField(max_length=50)),
                ('report_datetime', models.DateTimeField()),
                ('model', models.CharField(max_length=250)),
                ('field_name', models.CharField(max_length=50)),
                ('datatype', models.CharField(max_length=50)),
                ('value_str', models.CharField(max_length=50, null=True)),
                ('value_int', models.IntegerField(null=True)),
                ('value_date', models.DateField(null=True)),
                ('value_datetime', models.DateTimeField(null=True)),
            ],
            options={
                'ordering': ('identifier', 'report_datetime'),
            },
        ),
        migrations.AddIndex(
            model_name='reference',
            index=models.Index(fields=['identifier', 'timepoint', 'model'], name='edc_referen_identif_6df120_idx'),
        ),
        migrations.AddIndex(
            model_name='reference',
            index=models.Index(fields=['identifier', 'report_datetime', 'timepoint', 'model'], name='edc_referen_identif_351847_idx'),
        ),
        migrations.AddIndex(
            model_name='reference',
            index=models.Index(fields=['report_datetime', 'timepoint'], name='edc_referen_report__69deaa_idx'),
        ),
        migrations.AlterUniqueTogether(
            name='reference',
            unique_together={('identifier', 'timepoint', 'report_datetime', 'model', 'field_name')},
        ),
        migrations.AlterIndexTogether(
            name='reference',
            index_together={('identifier', 'timepoint', 'report_datetime', 'model', 'field_name')},
        ),
    ]



# needed for site_reference_configs.autodiscover test
from ..reference_model_config import ReferenceModelConfig
from ..site import site_reference_configs

reference = ReferenceModelConfig(
    fields=['f1', 'f2', 'f3', 'f4'], name='edc_reference.erik')

site_reference_configs.register(reference=reference)



from django.test import TestCase, tag

from ..refsets import Fieldset, FieldsetError


class DummyRefset:
    def __init__(self, i):
        self.f1 = i
        self.f2 = 'filter'[i]


class TestFieldset(TestCase):

    def setUp(self):
        refsets = []
        for i in range(0, 5):
            refsets.append(DummyRefset(i))
        self.fieldset = Fieldset(field='f1', refsets=refsets)

    def test_all(self):
        self.assertEqual(
            list(self.fieldset.all()), [0, 1, 2, 3, 4])

    def test_all_chained_to_orderby(self):
        self.assertEqual(
            list(self.fieldset.all().order_by('-f1')), [4, 3, 2, 1, 0])

    def test_filter(self):
        self.assertEqual(
            list(self.fieldset.filter(2)), [2])

    def test_filter_none(self):
        self.assertEqual(
            list(self.fieldset.filter()), [0, 1, 2, 3, 4])

    def test_filter_chained_to_orderby(self):
        self.assertEqual(
            list(self.fieldset.filter(2, 3, 4).order_by('-f1')), [4, 3, 2])

    def test_orderby_nothing(self):
        self.fieldset._refsets = []
        self.assertEqual(
            list(self.fieldset.order_by()), [])

    def test_first_last(self):
        self.assertEqual(self.fieldset.first(), 0)
        self.assertEqual(self.fieldset.last(), 4)
        self.assertEqual(self.fieldset.first(2), 2)
        self.assertEqual(self.fieldset.last(2), 2)
        self.assertEqual(self.fieldset.first(999), None)
        self.assertEqual(self.fieldset.last(999), None)
        self.assertEqual(self.fieldset.order_by('-f1').first(), 4)
        self.assertEqual(self.fieldset.order_by('-f1').last(), 0)

    def test_order_asc(self):
        self.assertEqual(
            list(self.fieldset.all().order_by('f1')), [0, 1, 2, 3, 4])

    def test_order_desc(self):
        self.assertEqual(
            list(self.fieldset.all().order_by('-f1')), [4, 3, 2, 1, 0])

    def test_order_asc_by_f2(self):
        self.assertEqual(
            list(self.fieldset.all().order_by('f2')), [4, 0, 1, 2, 3])

    def test_order_desc_by_f2(self):
        self.assertEqual(
            list(self.fieldset.all().order_by('-f2')), [3, 2, 1, 0, 4])

    def test_order_raises(self):
        self.assertRaises(
            FieldsetError,
            self.fieldset.all().order_by, 'blah')

    def test_as_iterator(self):
        for index, value in enumerate(self.fieldset):
            self.assertEqual(index, value)

    def test_index_support(self):
        for i in range(0, 5):
            self.assertEqual(self.fieldset[i], i)

    def test_len_support(self):
        self.assertEqual(len(self.fieldset), 5)



from edc_visit_schedule.schedule import Schedule
from edc_visit_schedule.visit import FormsCollection, Crf, Visit
from edc_visit_schedule.visit_schedule import VisitSchedule

from dateutil.relativedelta import relativedelta


crfs = FormsCollection(
    Crf(show_order=1, model=f'edc_reference.crfone', required=True),
)


visit0 = Visit(
    code='1000',
    title='Day 1',
    timepoint=0,
    rbase=relativedelta(days=0),
    rlower=relativedelta(days=0),
    rupper=relativedelta(days=6),
    crfs=crfs,
    facility_name='default')

schedule = Schedule(
    name='schedule',
    onschedule_model='edc_reference.onschedule',
    offschedule_model='edc_reference.offschedule',
    appointment_model='edc_appointment.appointment',
    consent_model='edc_reference.subjectconsent')

schedule.add_visit(visit0)

visit_schedule = VisitSchedule(
    name='visit_schedule',
    offstudy_model='edc_reference.subjectoffstudy',
    death_report_model='edc_reference.deathreport')

visit_schedule.add_schedule(schedule)



from django.test import TestCase, tag
from edc_base.utils import get_utcnow

from ..models import Reference


class TestModel(TestCase):

    reference_model_cls = Reference
    subject_identifier = '11111'

    def test_model_update(self):
        reference = self.reference_model_cls.objects.create(
            model=f'edc_reference.testmodel',
            identifier=self.subject_identifier,
            report_datetime=get_utcnow(),
            field_name='field_name')
        reference.update_value(
            value='5', internal_type='CharField')
        self.assertEqual(
            self.reference_model_cls.objects.get(id=reference.id).value, '5')
        self.assertEqual(
            self.reference_model_cls.objects.get(id=reference.id).value_str, '5')



# Generated by Django 2.0.1 on 2018-01-13 10:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('edc_reference', '0002_reference_value_uuid'),
    ]

    operations = [
        migrations.AddField(
            model_name='reference',
            name='related_name',
            field=models.CharField(max_length=100, null=True),
        ),
    ]



class FieldsetError(Exception):
    pass


class Fieldset:
    def __init__(self, field=None, refsets=None):
        self._values = None
        self.field = field
        self._refsets = refsets
        self._filter_values = None
        self.all()

    def __repr__(self):
        return f'{self.__class__.__name__}(field={self.field}, refsets={self._refsets})'

    def __iter__(self):
        return iter(self.values)

    def __getitem__(self, i):
        return self._values[i]

    def __len__(self):
        return len(self._values)

    @property
    def values(self):
        return self._values

    def all(self):
        self._values = [getattr(t, self.field) for t in self._refsets]
        return self

    def filter(self, *values):
        if len(values) > 0:
            self._values = [v for v in self._values if v in values]
            self._filter_values = values
        return self

    def order_by(self, field=None):
        """Re-order the collection ref objects by a single field
        and rebuild the values list.
        """
        field = field or 'report_datetime'
        self.ordering = field
        reverse = False
        if field.startswith('-'):
            field = field[1:]
            reverse = True
        try:
            getattr(self._refsets[0], field)
        except AttributeError as e:
            raise FieldsetError(
                f'Invalid ordering field. field={field}. Got {e}')
        except IndexError:
            pass
        else:
            self._refsets.sort(
                key=lambda x: getattr(x, field) or 0, reverse=reverse)
        self.all()
        if self._filter_values:
            self.filter(*self._filter_values)
        return self

    def first(self, value=None):
        """Returns the first value from the list of values.

        Excludes None.
        """
        if value:
            values = [v for v in self._values if v == value]
        else:
            values = [v for v in self._values if v is not None]
        try:
            return values[0]
        except IndexError:
            return None

    def last(self, value=None):
        """Returns the last value from the list of values.

        Excludes None.
        """
        if value:
            values = [v for v in self.values if v == value][-1:]
        else:
            values = [v for v in self._values if v is not None][-1:]
            # values = self.values[-1:]
        try:
            return values[0]
        except IndexError:
            return None



from django.core.exceptions import ObjectDoesNotExist
from edc_reference.site import SiteReferenceConfigError

from ..site import site_reference_configs


class RefsetError(Exception):
    pass


class RefsetOverlappingField(Exception):
    pass


class Refset:
    """An class that represents a queryset of a subject's references for a
    timepoint as a single object.
    """

    ordering_attrs = ['report_datetime', 'timepoint']

    def __init__(self, name=None, subject_identifier=None, report_datetime=None,
                 timepoint=None, reference_model_cls=None):
        # checking for these values so that field values
        # that are None are so because the reference instance
        # does not exist and not because these values were none.
        self._fields = []
        if not report_datetime:
            raise RefsetError('Expected report_datetime. Got None')
        if not timepoint:
            raise RefsetError('Expected timepoint. Got None')
        if not subject_identifier:
            raise RefsetError('Expected subject_identifier. Got None')
        self.subject_identifier = subject_identifier
        self.report_datetime = report_datetime
        self.timepoint = timepoint
        self.name = name

        self.model = '.'.join(name.split('.')[:2])

        opts = dict(
            identifier=self.subject_identifier,
            timepoint=timepoint,
            model=self.name)
        try:
            self._fields = dict.fromkeys(
                site_reference_configs.get_fields(name=self.name))
        except SiteReferenceConfigError as e:
            raise RefsetError(f'{e}. See {repr(self)}')
        try:
            self._fields.pop('report_datetime')
        except KeyError:
            pass
        try:
            references = reference_model_cls.objects.filter(**opts)
        except AttributeError as e:
            raise RefsetError(e)
        self._update_fields(references=references)

    def _update_fields(self, references=None):
        """Gets and updates each reference model instance for
        each field.
        """
        if references.count() == 0:
            self._fields.update(report_datetime=None)
            for field_name in self._fields:
                self._fields.update({field_name: None})
                setattr(self, field_name, None)
        else:
            for field_name in self._fields:
                try:
                    obj = references.get(field_name=field_name)
                except ObjectDoesNotExist:
                    self._fields.update({field_name: None})
                else:
                    self._fields.update({field_name: obj.value})
            for key, value in self._fields.items():
                try:
                    existing_value = getattr(self, key)
                except AttributeError:
                    setattr(self, key, value)
                else:
                    if existing_value != value:
                        raise RefsetOverlappingField(
                            f'Attribute {key} already exists with a different value. '
                            f'Got {existing_value} == {value}. See {self.name}')
            self._fields.update(report_datetime=self.report_datetime)
            self._fields.update(visit_code=self.timepoint)

    @property
    def visit_code(self):
        return self.timepoint

    def __repr__(self):
        return (f'{self.__class__.__name__}(name={self.name},'
                f'subject_identifier={self.subject_identifier},'
                f'timepoint={self.timepoint}) <{[f for f in self._fields]}>')



from copy import copy
from django.apps import apps as django_apps

from .fieldset import Fieldset
from .refset import Refset


class LongitudinalRefsetError(Exception):
    pass


class InvalidOrdering(Exception):
    pass


class NoRefsetObjectsExist(Exception):
    pass


class LongitudinalRefset:
    """A collection of a subject's `Refset` objects for a given visit model.
    """

    fieldset_cls = Fieldset
    refset_cls = Refset

    def __init__(self, name=None, subject_identifier=None, visit_model=None,
                 reference_model_cls=None, **options):
        self.name = name
        self.model = '.'.join(name.split('.')[:2])
        self.ordering = None
        self.subject_identifier = subject_identifier
        try:
            reference_model_cls = django_apps.get_model(reference_model_cls)
        except AttributeError:
            pass
        opts = dict(
            identifier=self.subject_identifier,
            model=visit_model,
            field_name='report_datetime',
            **options)
        try:
            self.visit_references = reference_model_cls.objects.filter(**opts)
        except ValueError as e:
            raise LongitudinalRefsetError(
                f'{e}. name={self.name}. Got {opts}.')
        self._refsets = []
        for visit_reference in self.visit_references:
            self._refsets.append(
                self.refset_cls(
                    name=self.name,
                    subject_identifier=subject_identifier,
                    report_datetime=visit_reference.report_datetime,
                    timepoint=visit_reference.timepoint,
                    reference_model_cls=reference_model_cls))
        self.ordering_attrs = copy(self.refset_cls.ordering_attrs)
        for refset in self._refsets:
            self.ordering_attrs.extend(list(refset._fields))
        self.ordering_attrs = list(set(self.ordering_attrs))
        self.order_by('report_datetime')

    def __repr__(self):
        return f'{self.__class__.__name__}({self._refsets})'

    def __iter__(self):
        return iter(self._refsets)

    def __getitem__(self, i):
        return self._refsets[i]

    def order_by(self, field=None):
        """Re-order the collection ref objects by a single field.
        """
        if field and field.replace('-', '') not in self.ordering_attrs:
            raise InvalidOrdering(
                f'Invalid ordering field. field={field}. Expected one of '
                f'{self.ordering_attrs}')
        field = field or 'report_datetime'
        self.ordering = field
        reverse = False
        if field.startswith('-'):
            field = field[1:]
            reverse = True
        try:
            self._refsets.sort(
                key=lambda x: getattr(x, field), reverse=reverse)
        except TypeError:
            null_refsets = [
                x for x in self._refsets if getattr(x, field) is None]
            notnull_refsets = [
                x for x in self._refsets if getattr(x, field) is not None]
            notnull_refsets.sort(
                key=lambda x: getattr(x, field), reverse=reverse)
            self._refsets = notnull_refsets + null_refsets
        return self

    def fieldset(self, field_name=None):
        """Returns a FieldSet for this field name.
        """
        if not self._refsets:
            raise NoRefsetObjectsExist(
                f'No Ref objects exist for this subject. '
                f'Got subject_identifier={self.subject_identifier}')
        return self.fieldset_cls(field=field_name, refsets=self._refsets)



from dateutil.relativedelta import relativedelta
from django.test import TestCase, tag

from edc_base.utils import get_utcnow

from ..refsets import LongitudinalRefset, InvalidOrdering, NoRefsetObjectsExist
from ..models import Reference
from ..reference_model_config import ReferenceModelConfig
from ..site import site_reference_configs
from .models import SubjectVisit, CrfOne


class TestLongitudinal(TestCase):

    def setUp(self):
        self.subject_identifier = '12345'
        site_reference_configs.registry = {}
        site_reference_configs.loaded = False
        reference = ReferenceModelConfig(
            name='edc_reference.subjectvisit',
            fields=['report_datetime', 'visit_code'])
        site_reference_configs.register(reference=reference)
        reference = ReferenceModelConfig(
            name='edc_reference.crfone',
            fields=['field_date', 'field_datetime', 'field_int', 'field_str'])
        site_reference_configs.register(reference=reference)

        values = [
            ('NEG', get_utcnow() - relativedelta(years=3)),
            ('POS', get_utcnow() - relativedelta(years=2)),
            ('POS', get_utcnow() - relativedelta(years=1)),
        ]
        for index in [1, 2, 3]:
            SubjectVisit.objects.create(
                subject_identifier=self.subject_identifier,
                report_datetime=get_utcnow() - relativedelta(months=index),
                visit_code=str(index))
        for index, subject_visit in enumerate(SubjectVisit.objects.filter(
                subject_identifier=self.subject_identifier).order_by('report_datetime')):
            CrfOne.objects.create(
                subject_visit=subject_visit,
                field_str=values[index][0],
                field_datetime=values[index][1])

    def test_longitudinal_refset(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        self.assertEqual([ref.timepoint for ref in refset], ['3', '2', '1'])

    def test_longitudinal_refset_uses_subject_visit_report_datetime(self):
        longitudinal_refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        subject_visits = SubjectVisit.objects.filter(
            subject_identifier=self.subject_identifier).order_by('report_datetime')
        report_datetimes = [obj.report_datetime for obj in subject_visits]
        self.assertEqual(
            [ref.report_datetime for ref in longitudinal_refset], report_datetimes)
        self.assertEqual(
            [v.report_datetime for v in longitudinal_refset.visit_references],
            report_datetimes)

    def test_no_refsets(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        refset._refsets = []
        self.assertRaises(
            NoRefsetObjectsExist,
            refset.fieldset, 'field_name')

    def test_ordering(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference).order_by('-report_datetime')
        self.assertEqual(
            [ref.timepoint for ref in refset], ['1', '2', '3'])
        refset.order_by()
        self.assertEqual([ref.timepoint for ref in refset], ['3', '2', '1'])

    def test_bad_ordering(self):
        self.assertRaises(
            InvalidOrdering,
            LongitudinalRefset(
                subject_identifier=self.subject_identifier,
                visit_model='edc_reference.subjectvisit',
                name='edc_reference.crfone',
                reference_model_cls=Reference).order_by, 'blah')

    def test_get(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        self.assertEqual(
            refset.fieldset('field_str').all().values, ['NEG', 'POS', 'POS'])
        self.assertEqual(
            refset.fieldset('field_str').all().order_by(
                '-report_datetime').values,
            ['POS', 'POS', 'NEG'])

    def test_get2(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        self.assertEqual(
            refset.fieldset('field_str').all().order_by(
                'field_datetime').values,
            ['NEG', 'POS', 'POS'])
        self.assertEqual(
            refset.fieldset('field_str').order_by(
                '-field_datetime').all().values,
            ['POS', 'POS', 'NEG'])

    def test_get_last(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        self.assertEqual(refset.fieldset('field_str').order_by(
            'field_datetime').last(), 'POS')
        self.assertEqual(
            refset.fieldset('field_str').order_by('-field_datetime').last(), 'NEG')

    def test_repr(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls=Reference)
        self.assertTrue(repr(refset))
        for ref in refset:
            self.assertTrue(repr(ref))

    def test_with_model_name(self):
        refset = LongitudinalRefset(
            subject_identifier=self.subject_identifier,
            visit_model='edc_reference.subjectvisit',
            name='edc_reference.crfone',
            reference_model_cls='edc_reference.reference')
        self.assertTrue(repr(refset))
        for ref in refset:
            self.assertTrue(repr(ref))



from django.contrib.auth.decorators import login_required
from django.utils.decorators import method_decorator
from django.views.generic.base import TemplateView
from edc_base.view_mixins import EdcBaseViewMixin
from edc_navbar import NavbarViewMixin


class HomeView(EdcBaseViewMixin, NavbarViewMixin, TemplateView):

    template_name = 'edc_reference/home.html'
    navbar_name = 'edc_reference'
    navbar_selected_item = 'reference'

    @method_decorator(login_required)
    def dispatch(self, *args, **kwargs):
        return super().dispatch(*args, **kwargs)



import arrow
import sys

from django.apps import apps as django_apps
from edc_reference.models import Reference

from .reference import ReferenceUpdater
from .site import site_reference_configs


class PopulaterAttributeError(Exception):
    pass


class DryRunDummy:
    def __init__(self, **kwargs):
        pass


class Populater:

    reference_updater_cls = ReferenceUpdater

    def __init__(self, names=None, exclude_names=None, skip_existing=None,
                 dry_run=None, delete_existing=None):
        self.skip_existing = skip_existing
        self.delete_existing = delete_existing
        if not names:
            names = list(site_reference_configs.registry)
        if not exclude_names:
            exclude_names = []
        exclude_names = [n.strip() for n in exclude_names]
        self.names = [n.strip() for n in names if n not in exclude_names]
        self.dry_run = dry_run

    def summarize(self):
        for name in self.names:
            reference_model = site_reference_configs.get_reference_model(
                name=name)
            reference_model_cls = django_apps.get_model(reference_model)
            count = reference_model_cls.objects.filter(model=name).count()
            sys.stdout.write(f' * {name}: {count} records\n')

    def populate(self):
        if self.dry_run:
            self.reference_updater_cls = DryRunDummy
        t_start = arrow.utcnow().to('Africa/Gaborone').strftime('%H:%M')
        sys.stdout.write(
            f'Populating reference model. Started: {t_start}\n')
        sys.stdout.write(
            f' - found {len(site_reference_configs.registry)} reference names in registry.\n')
        sys.stdout.write(
            f' - running for {len(self.names)} selected reference names.\n')
        if self.skip_existing:
            sys.stdout.write(
                f' - skipping reference names with existing references\n')
        if self.dry_run:
            sys.stdout.write(
                f' - This is a dry run. No data will be created/modified.\n')

        names = [name for name in self.names if not self.skip(name=name)]

        for name in names:
            site_reference_configs.get_config(name=name)

        sys.stdout.write(f' * models are {names}    \n')

        if self.delete_existing:
            sys.stdout.write(f' * deleting existing records ... \r')
            if not self.dry_run:
                for name in names:
                    Reference.objects.filter(model=name).delete()
            sys.stdout.write(f' * deleting existing records ... done.\n')

        for name in names:
            index = 0
            sys.stdout.write(f' * {name}           \r')
            model_cls = django_apps.get_model('.'.join(name.split('.')[:2]))
            qs = model_cls.objects.all()
            total = qs.count()
            sub_start_time = arrow.utcnow().to('Africa/Gaborone').datetime
            for index, model_obj in enumerate(qs):
                index += 1
                sub_end_time = arrow.utcnow().to('Africa/Gaborone').datetime
                tdelta = sub_end_time - sub_start_time
                sys.stdout.write(
                    f' * {name} {index} / {total} ... {str(tdelta)}    \r')
                self.reference_updater_cls(model_obj=model_obj)
            sub_end_time = arrow.utcnow().to('Africa/Gaborone').datetime
            tdelta = sub_end_time - sub_start_time
            sys.stdout.write(
                f' * {name} {index} / {total} . OK  in {str(tdelta)}      \n')
        t_end = arrow.utcnow().to('Africa/Gaborone').strftime('%H:%M')
        sys.stdout.write(f'Done. Ended: {t_end}\n')

    def skip(self, name=None):
        if self.skip_existing:
            reference_model = site_reference_configs.get_reference_model(
                name=name)
            reference_model_cls = django_apps.get_model(
                reference_model)
            return reference_model_cls.objects.filter(model=name).exists()
        return False



from django.db import models
from django.core.exceptions import ObjectDoesNotExist


class ReferenceManager(models.Manager):

    def get_by_natural_key(self, identifier, timepoint, report_datetime,
                           model, field_name):
        return self.get(
            identifier=identifier,
            timepoint=timepoint,
            report_datetime=report_datetime,
            model=model,
            field_name=field_name)

    def filter_crf_for_visit(self, name=None, visit=None):
        """Returns a queryset of reference model instances
        for this model on this visit.
        """
        return self.filter(
            identifier=visit.subject_identifier,
            model=name,
            report_datetime=visit.report_datetime,
            timepoint=visit.visit_code)

    def get_crf_for_visit(self, name=None, visit=None, field_name=None):
        """Returns an instance of reference model
        for this model on this visit for this field.
        """
        try:
            model_obj = self.get(
                identifier=visit.subject_identifier,
                model=name,
                report_datetime=visit.report_datetime,
                timepoint=visit.visit_code,
                field_name=field_name)
        except ObjectDoesNotExist:
            model_obj = None
        return model_obj

    def get_requisition_for_visit(self, name=None, visit=None):
        """Returns an instance of reference model
        for this requisition on this visit for this panel.
        """
        opts = dict(
            identifier=visit.subject_identifier,
            model=name,
            report_datetime=visit.report_datetime,
            timepoint=visit.visit_code,
            field_name='panel')
        try:
            model_obj = self.get(**opts)
        except ObjectDoesNotExist:
            model_obj = None
        return model_obj



from django.db import models
from django.db.models.deletion import PROTECT
from edc_base.model_mixins import BaseUuidModel
from edc_base.utils import get_utcnow
from edc_lab.models.model_mixins import PanelModelMixin

from ..model_mixins import ReferenceModelMixin, RequisitionReferenceModelMixin


class CrfModelMixin(models.Model):

    @property
    def visit(self):
        return self.subject_visit

    class Meta:
        abstract = True


class SubjectVisit(ReferenceModelMixin, BaseUuidModel):

    subject_identifier = models.CharField(max_length=50)

    report_datetime = models.DateTimeField(default=get_utcnow)

    visit_code = models.CharField(max_length=50)


class SubjectRequisition(CrfModelMixin, PanelModelMixin, RequisitionReferenceModelMixin,
                         BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    panel_name = models.CharField(max_length=50)


class TestModel(CrfModelMixin, ReferenceModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    field_str = models.CharField(max_length=50)


class TestModelBad(CrfModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    field_str = models.CharField(max_length=50)


class CrfOne(CrfModelMixin, ReferenceModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    field_str = models.CharField(max_length=50)

    field_date = models.DateField(null=True)

    field_datetime = models.DateTimeField(null=True)

    field_int = models.IntegerField(null=True)


class CrfWithBadField(CrfModelMixin, ReferenceModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    field_str = models.CharField(max_length=50)

    field_date = models.DateField(null=True)

    field_datetime = models.DateTimeField(null=True)

    field_int = models.IntegerField(null=True)


class CrfWithDuplicateField(CrfModelMixin, ReferenceModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    field_str = models.CharField(max_length=50)

    field_date = models.DateField(null=True)

    field_datetime = models.DateTimeField(null=True)

    field_int = models.IntegerField(null=True)


class CrfWithUnknownDatatype(CrfModelMixin, ReferenceModelMixin, BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=PROTECT)

    report_datetime = models.DateTimeField(default=get_utcnow)

    field_decimal = models.DecimalField(decimal_places=2, max_digits=10)



from django.apps import apps as django_apps
from django.core.exceptions import ObjectDoesNotExist

from ..site import site_reference_configs


class ReferenceObjectDoesNotExist(Exception):
    pass


class ReferenceGetter:
    """A class that gets the reference model instance for a given
    model or attributes of the model.

    See also ReferenceModelMixin.
    """

    def __init__(self, name=None, field_name=None, model_obj=None, visit_obj=None,
                 subject_identifier=None, report_datetime=None, visit_code=None,
                 create=None):
        self._object = None
        self.value = None
        self.has_value = False
        self.field_name = field_name

        if model_obj:
            try:
                # given a crf model as model_obj
                self.report_datetime = model_obj.visit.report_datetime
                self.subject_identifier = model_obj.visit.subject_identifier
                self.visit_code = model_obj.visit.visit_code
                self.name = model_obj.reference_name
            except AttributeError:
                # given a visit model as model_obj
                self.subject_identifier = model_obj.subject_identifier
                self.report_datetime = model_obj.report_datetime
                self.visit_code = model_obj.visit_code
                self.name = model_obj.reference_name
        elif visit_obj:
            self.name = name
            self.subject_identifier = visit_obj.subject_identifier
            self.report_datetime = visit_obj.report_datetime
            self.visit_code = visit_obj.visit_code
        else:
            # given only the attrs
            self.name = name
            self.subject_identifier = subject_identifier
            self.report_datetime = report_datetime
            self.visit_code = visit_code
        reference_model = site_reference_configs.get_reference_model(
            name=self.name)
        reference_model_cls = django_apps.get_model(reference_model)
        try:
            self.object = reference_model_cls.objects.get(**self._options)
        except ObjectDoesNotExist as e:
            if create:
                self.object = reference_model_cls.objects.create(
                    **self._options)
                # note: updater needs to "update_value"
            else:
                raise ReferenceObjectDoesNotExist(
                    f'{e}. Using {self._options}')
        else:
            self.value = getattr(self.object, 'value')
            self.has_value = True
        setattr(self, self.field_name, self.value)

    def __repr__(self):
        return (f'<{self.__class__.__name__}({self.name}.{self.field_name}\','
                f'\'{self.subject_identifier},{self.report_datetime}'
                f') value={self.value}, has_value={self.has_value}>')

    @property
    def _options(self):
        return dict(
            identifier=self.subject_identifier,
            model=self.name,
            report_datetime=self.report_datetime,
            timepoint=self.visit_code,
            field_name=self.field_name)



from django.contrib import admin

from .admin_site import edc_reference_admin
from .models import Reference


@admin.register(Reference, site=edc_reference_admin)
class ReferenceAdmin(admin.ModelAdmin):

    list_display = (
        'identifier', 'model', 'report_datetime', 'timepoint', 'field_name')
    list_filter = ('model', 'timepoint', 'field_name')
    search_fields = ('identifier', )



from django.core.management.base import BaseCommand
from edc_reference.site import site_reference_configs


class Command(BaseCommand):

    help = 'Validates the reference model config'

    def handle(self, *args, **options):
        site_reference_configs.validate()



from django.apps import apps as django_apps


class ReferenceModelValidationError(Exception):
    pass


class ReferenceFieldValidationError(Exception):
    pass


class ReferenceDuplicateField(Exception):
    pass


class ReferenceFieldAlreadyAdded(Exception):
    pass


class ReferenceModelConfig:

    reference_model = 'edc_reference.reference'

    def __init__(self, name=None, fields=None):
        """
        Keywords:
            name = app_label.model_name for CRFs
            name = app_label.model_name.panel for Requisitions
        """

        if not fields:
            raise ReferenceFieldValidationError('No fields declared.')
        self.field_names = list(set(fields))
        self.field_names.sort()
        self.name = name.lower()
        self.model = '.'.join(name.split('.')[:2])

        if len(fields) != len(self.field_names):
            raise ReferenceDuplicateField(
                f'Duplicate field detected. Got {fields}. See \'{self.name}\'')

    def add_fields(self, fields=None):
        for field_name in fields:
            if field_name in self.field_names:
                raise ReferenceFieldAlreadyAdded(
                    f'Field already added. Got {field_name}. See \'{self.name}\'')
        self.field_names.extend(fields)
        self.field_names = list(set(self.field_names))
        self.field_names.sort()

    def __repr__(self):
        return f'{self.__class__.__name__}(name={self.name}, fields={self.field_names})'

    def check(self):
        """Validates the model class by doing a django.get_model lookup
        and confirming all fields exist on the model class.
        """
        try:
            model_cls = django_apps.get_model(self.model)
        except LookupError:
            raise ReferenceModelValidationError(
                f'Invalid app label or model name. Got {self.model}. See {repr(self)}.')
        model_field_names = [fld.name for fld in model_cls._meta.get_fields()]
        for field_name in self.field_names:
            if field_name not in model_field_names:
                raise ReferenceFieldValidationError(
                    f'Invalid reference field. Got {field_name} not found '
                    f'on model {repr(model_cls)}. See {repr(self)}.')
        try:
            model_cls.reference_updater_cls
            model_cls.reference_deleter_cls
        except AttributeError:
            raise ReferenceFieldValidationError(
                f'Missing reference model mixin. See model {repr(model_cls)}')
